(*
MIDI unit
by DooMeeR (webmaster@doomeer.com)

Implement MIDI-related classes and methods:
- Variable-Length Value type, for delta-time (weird integer type with variable size, used in MIDI file format)
- MIDI file format use
- MIDI output
*)

(*
Midi notes:

       || Note Numbers
Octave || C   | C#  | D   | D#  | E   | F   | F#  | G   | G#  | A   | A#  | B
       || Do  | Do# | R  | R# | Mi  | Fa  | Fa# | Sol | Sol#| La  | La# | Si
-------++-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----
     0 || 0   | 1   | 2   | 3   | 4   | 5   | 6   | 7   | 8   | 9   | 10  | 11
     1 || 12  | 13  | 14  | 15  | 16  | 17  | 18  | 19  | 20  | 21  | 22  | 23
     2 || 24  | 25  | 26  | 27  | 28  | 29  | 30  | 31  | 32  | 33  | 34  | 35
     3 || 36  | 37  | 38  | 39  | 40  | 41  | 42  | 43  | 44  | 45  | 46  | 47
     4 || 48  | 49  | 50  | 51  | 52  | 53  | 54  | 55  | 56  | 57  | 58  | 59
     5 || 60  | 61  | 62  | 63  | 64  | 65  | 66  | 67  | 68  | 69  | 70  | 71
     6 || 72  | 73  | 74  | 75  | 76  | 77  | 78  | 79  | 80  | 81  | 82  | 83
     7 || 84  | 85  | 86  | 87  | 88  | 89  | 90  | 91  | 92  | 93  | 94  | 95
     8 || 96  | 97  | 98  | 99  | 100 | 101 | 102 | 103 | 104 | 105 | 106 | 107
     9 || 108 | 109 | 110 | 111 | 112 | 113 | 114 | 115 | 116 | 117 | 118 | 119
    10 || 120 | 121 | 122 | 123 | 124 | 125 | 126 | 127 |
*)
unit Midi;

interface

uses Classes, Dialogs, SysUtils, MMSystem, Windows; // MMSystem contains a lot of constants and types used for MIDI APIs calls

type
  Array8 = array[1..8] of byte; // Used when you got a vlv in... non-vlv format
  vlv = class // Variable-Length Value (used for delta-times)
    Value: array of byte;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure SetFromInt(int: integer); // Set from normal int
    procedure SetFromVLVA8(A8: Array8); // Set from an Array8 VLV
    function Size: integer; // Nomber of bytes of Value (equal Length(Value))
    function IntVal: integer; // Get Value in a normal integer form
    function CardVal: cardinal; // Get Value in a normal unsigned integer form
  end;

  // TMidiFormat is used in the MIDI header. See MIDI file format for more details.
  TMidiFormat = (mfSingle, mfSync, mfAsync);
  // TMidi is used to load and save MIDI files easily, separating tracks, and events in tracks
  TMidi = class
    Tracks: TList;
    Format: TMidiFormat;
    TicksPerQNote: word; // Delta-Time ticks per quarter note
    constructor Create; virtual;
    destructor Destroy; override;
    procedure SaveToFile(filename: string); // Save to midi file
    procedure SaveToStream(s: TStream); // Save to a stream in midi format (as in a file)
    procedure LoadFromFile(filename: string); // Load from midi file
    procedure LoadFromStream(s: TStream); // Load from a stream in midi format (as in a file)
  end;
  // TMidiTrack contains a TList of MIDI events
  TMidiTrack = class
    Events: TList;
    constructor Create; virtual;
    destructor Destroy; override;
    function SaveSize: integer; // Size taken when saving, not including header
    procedure SaveToStream(s: TStream); // Save track in a stream, with header and events. Used by TMidi.SaveToStream
    procedure LoadFromStream(s: TStream); // Load track from a stream, with header and events. Used by TMidi.LoadFromStream
  end;
  // TMidiEvent is the base class for MIDI events
  TMidiEvent = class
    Time: vlv; // Delta-Time (when the event is called)
    Num: byte; // Event "name"
    Meta: boolean; // Meta-Event? (other kind of event)
    Channel: byte; // Channel on which the event is used (not used for Meta-Events)
    constructor Create; virtual;
    destructor Destroy; override;
    function SaveSize: byte; virtual; abstract; // Size taken when saving
    procedure SaveToStream(s: TStream); virtual; // Save basic event information (time, num, meta, channel). Used by TMidiTrack.SaveToStream
    procedure LoadFromStream(s: TStream); virtual; abstract; // Implemented by children to load parameters if needed
  end;
  // Midi event exemple:
  {TMidiImplementedEvent = class (TMidiEvent)
    constructor Create; override;
    function SaveSize: byte; override; // Return the size taken when saving the event, including delta-time size. Used by TMidiTrack.SaveSize. Implementation of the abstract function
    procedure SaveToStream(s: TStream); override; // Save event in a stream. Use inherited TMidiEvent.SaveToStream, and add other event-specific informations.
    procedure LoadFromStream(s: TStream); override; // Load parameters (not num nor channel nor meta)
  end;}
  TMidiEndOfTrackEvent = class (TMidiEvent) // Don't use it - automatically added when saving.
    constructor Create; override;
    function SaveSize: byte; override;
    procedure SaveToStream(s: TStream); override;
    procedure LoadFromStream(s: TStream); override;
  end;
  TMidiNoteONEvent = class (TMidiEvent) // Activate a note
    Note, Velocity: byte;
    constructor Create; override;
    function SaveSize: byte; override;
    procedure SaveToStream(s: TStream); override;
    procedure LoadFromStream(s: TStream); override;
  end;
  TMidiNoteOFFEvent = class (TMidiEvent) // Deactivate a note
    Note: byte;
    constructor Create; override;
    function SaveSize: byte; override;
    procedure SaveToStream(s: TStream); override;
    procedure LoadFromStream(s: TStream); override;
  end;
  TMidiChangePatchEvent = class (TMidiEvent) // Change patch (instrument)
    Patch: byte;
    constructor Create; override;
    function SaveSize: byte; override;
    procedure SaveToStream(s: TStream); override;
    procedure LoadFromStream(s: TStream); override;
  end;

  // TMidiOutput able you to open a midi device for output, and send it midi events in order to play midi music
  // I couldn''t test the volume properties, i got the "function isn't supported" error message, so try them and tell me if it works :)
  TMidiOutput = class
  private
    function GetVolume: word;
    function GetVolumeLeft: word;
    function GetVolumeRight: word;
    procedure SetVolume(const Value: word);
    procedure SetVolumeRight(const Value: word);
    procedure SetVolumeLeft(Value: word);
    procedure SetDeviceID(const Value: integer);
  protected
    // Use public properties to access these values (they should be modified only by the class)
    FHandle: HMIDIOUT;
    FOpened: boolean;
    FDeviceID: integer;
  public
    // Methods you can use
    constructor Create(ffDeviceID: integer; OpenNow: boolean); // Enter Device num (most used: 0). OpenNow automatically call the Open method.
    destructor Destroy; override;
    procedure Open; // Open device for output
    procedure Close; // Close device
    procedure ShortMsg(msg: cardinal); // Send midi information
    procedure NoteON(channel, note, volume: byte); // Activate a note
    procedure NoteOFF(channel, note: byte); // Deactivate a note
    procedure Drum(note, volume: byte); // NoteON on channel 9 (doesn't need NoteOFF)
    procedure ChangePatch(channel, patch: byte); // Change patch (instrument)
    procedure NotesOFF(channel: byte); // Deactivate all notes on a channel

    // Properties you can use
    property Handle: HMIDIOUT read FHandle; // Handle returned by midiOutOpen in the Open method, used in every calls to MIDI APIs (read-only)
    property Opened: boolean read FOpened; // Is MIDI output opened? (read-only)
    property DeviceID: integer read FDeviceID write SetDeviceID; // DeviceID set in the constructor. Changing it will close midi output
    property VolumeLeft: word read GetVolumeLeft write SetVolumeLeft;    // Volume left
    property VolumeRight: word read GetVolumeRight write SetVolumeRight; // Volume right
    property VolumeMono: word read GetVolumeLeft write SetVolumeLeft;    // Volume if device doesn't support stereo
    property Volume: word read GetVolume write SetVolume;                // General volume = (volumeleft + volumeright) / 2
  end;

function GetMidiErrorText(err: integer): string; // Midi functions always return values: 0 if no error, more if an error has occured. Use this function to translate the number in text
function Invert4B(bytes: integer): integer; // Invert an integer for saving in midi format (1 2 3 4 instead of 4 3 2 1)
function Invert2B(bytes: word): word; // Invert a word for saving in midi format (1 2 instead of 2 1)

implementation

{ Global }

function GetMidiErrorText(err: integer): string;
begin
  case err of
    MMSYSERR_NOERROR:      result:='No error'; // 0
    MMSYSERR_ERROR:        result:='Unspecified error'; // 1
    MMSYSERR_BADDEVICEID:  result:='Device ID out of range'; // 2
    MMSYSERR_NOTENABLED:   result:='Driver failed enable'; // 3
    MMSYSERR_ALLOCATED:    result:='Device already allocated'; // 4
    MMSYSERR_INVALHANDLE:  result:='Device handle is invalid'; // 5
    MMSYSERR_NODRIVER:     result:='No device driver present'; // 6
    MMSYSERR_NOMEM:        result:='Memory allocation error'; // 7
    MMSYSERR_NOTSUPPORTED: result:='Function isn''t supported'; // 8
    MMSYSERR_BADERRNUM:    result:='Error value out of range'; // 9
    MMSYSERR_INVALFLAG:    result:='Invalid flag passed'; // 10
    MMSYSERR_INVALPARAM:   result:='Invalid parameter passed'; // 11
    MMSYSERR_HANDLEBUSY:   result:='Handle being used simultaneously on another thread (eg callback)'; // 12
    MMSYSERR_INVALIDALIAS: result:='Specified alias not found'; // 13
    MMSYSERR_BADDB:        result:='Bad registry database'; // 14
    MMSYSERR_KEYNOTFOUND:  result:='Registry key not found'; // 15
    MMSYSERR_READERROR:    result:='Registry read error'; // 16
    MMSYSERR_WRITEERROR:   result:='Registry write error'; // 17
    MMSYSERR_DELETEERROR:  result:='Registry delete error'; // 18
    MMSYSERR_VALNOTFOUND:  result:='Registry value not found'; // 19
    MMSYSERR_NODRIVERCB:   result:='Driver does not call DriverCallback'; // 20
  else
    result:='Unknown error';
  end;
end;

function Invert4B(bytes: integer): integer;
var a, b, c, d: byte;
begin
  a:=bytes and $FF;
  b:=(bytes and $FF00) shr 8;
  c:=(bytes and $FF0000) shr 16;
  d:=(bytes and $FF000000) shr 24;
  result:=d+(c shl 8)+(b shl 16)+(a shl 24); 
end;

function Invert2B(bytes: word): word;
var a, b: byte;
begin
  a:=bytes and $FF;
  b:=(bytes and $FF00) shr 8;
  result:=b+(a shl 8);
end;

{ vlv }

constructor vlv.Create;
begin
  SetLength(Value, 1);
  Value[0]:=0;
end;

destructor vlv.Destroy;
begin
  SetLength(Value, 0);
end;

function vlv.Size: integer;
begin
  result:=length(Value);
end;

procedure vlv.SetFromInt(int: integer);
var a, b, c, d, e: byte;
    tval: array of byte;
begin
  a:=int and 127;
  b:=(int and (127 shl 7)) shr 7;
  c:=(int and (127 shl 14)) shr 14;
  d:=(int and (127 shl 21)) shr 21;
  e:=(int and (127 shl 28)) shr 28;
  SetLength(tval, 1);
  tval[0]:=a+128;
  if b>0 then
  begin
    SetLength(tval, 2);
    tval[1]:=b+128;
    if c>0 then
    begin
      SetLength(tval, 3);
      tval[2]:=c+128;
      if d>0 then
      begin
        SetLength(tval, 4);
        tval[3]:=d+128;
        if e>0 then
        begin
          SetLength(tval, 5);
          tval[5]:=e+128;
        end;
      end;
    end;
  end;
  SetLength(Value, Length(tval));
  for a:=0 to High(tval) do Value[a]:=tval[High(tval)-a];
  Value[High(Value)]:=Value[High(Value)]-128;
end;

procedure vlv.SetFromVLVA8(A8: Array8);
var i: integer;
begin
  SetLength(Value, 0);
  for i:=1 to 8 do
  begin
    SetLength(Value, length(Value)+1);
    Value[High(Value)]:=A8[i];
    if A8[i]<128 then break;
  end;
end;

function vlv.CardVal: cardinal;
var i, h: byte;
begin
  result:=0;
  h:=High(Value);
  for i:=0 to h do
  begin
    if Value[i]>=128 then result:=result or ((Value[i] and 127) shl ((h-i)*7))
    else begin
      result:=result or ((Value[i]) shl ((h-i)*7));
      if i<h then showmessage('Weird thing hapenned: '+inttostr(i)+' / '+inttostr(h));
      break;
    end;
    if i>4 then break; // integers can't use more than 4 bytes
  end;
end;

function vlv.IntVal: integer;
var i, h: byte;
begin
  result:=0;
  h:=High(Value);
  for i:=0 to h do
  begin
    if Value[i]>=128 then result:=result or ((Value[i] and 127) shl ((h-i)*7))
    else begin
      result:=result or ((Value[i]) shl ((h-i)*7));
      if i<h then showmessage('Weird thing hapenned: '+inttostr(i)+' / '+inttostr(h));
      break;
    end;
    if i>4 then break; // integers can't use more than 4 bytes
  end;
end;

{ TMidiTrack }

constructor TMidiTrack.Create;
begin
  Events:=TList.Create;
end;

destructor TMidiTrack.Destroy;
var i: integer;
begin
  for i:=0 to Events.Count-1 do TMidiEvent(Events[i]).Free;
  inherited Destroy;
end;

procedure TMidiTrack.LoadFromStream(s: TStream);
var error: boolean;
    k: cardinal;
    n: integer;
    dt: Array8; // event delta-time in array8 format (for loading)
    b: byte;
    Event: TMidiEvent; // abstract event class
    eventdt: vlv; // event delta-time in vlv format
    kind, kind2: byte; // kind: FF if meta-event, or event num else. kind2: event num if meta
    chan: byte; // kind + chan = byte 1 if not a meta event...
    mesize: byte; // size of meta-event
begin
  error:=false;
  s.Read(k, 4);
  if k<>$6B72544D then error:=true; // MTrk
  s.Read(k, 4); // Size of track

  if error then raise Exception.Create('Invalid MIDI file format.') else
  begin
    // Events
    while s.Position<=s.Size do
    begin
      // Delta-Time
      n:=0;
      while true do
      begin
        s.Read(b, 1);
        if n<8 then
        begin
          inc(n);
          dt[n]:=b;
          if b<128 then break;
        end;
      end;
      eventdt:=vlv.Create;
      eventdt.SetFromVLVA8(dt);
      // Event kind
      s.Read(kind, 1);
      if kind=$FF then
      begin
        s.Read(kind2, 1);
        if kind2=$2F then // End of Track
        begin
          s.Read(mesize, 1);
          break;
        end;
      end else begin
        chan:=kind and $0F;
        kind:=kind and $F0;
        Event:=nil;
        if kind=$C0 then Event:=TMidiChangePatchEvent.Create else
        if kind=$80 then Event:=TMidiNoteOFFEvent.Create else
        if kind=$90 then Event:=TMidiNoteONEvent.Create;
        if Event<>nil then
        begin
          Events.Add(Event);
          Event.LoadFromStream(s);
          Event.Time:=eventdt;
          Event.Channel:=Chan;
        end else raise Exception.Create('Invalid MIDI file format.');
      end;
    end;
  end;
end;

function TMidiTrack.SaveSize: integer;
var i: integer;
begin
  result:=0;
  for i:=0 to Events.Count-1 do
    result:=result+TMidiEvent(Events[i]).SaveSize;
end;

procedure TMidiTrack.SaveToStream(s: TStream);
var i: integer;
begin
  Events.Add(TMidiEndOfTrackEvent.Create);
  s.Write('MTrk', 4);
  i:=Invert4B(SaveSize);
  s.Write(i, 4);
  for i:=0 to Events.Count-1 do
  begin
    TMidiEvent(Events[i]).SaveToStream(s);
  end;
  Events.Delete(Events.Count-1);
end;

{ TMidiEvent }

constructor TMidiEvent.Create;
begin
  Time:=vlv.Create;
end;

destructor TMidiEvent.Destroy;
begin
  Time.Free;
  inherited Destroy;
end;

procedure TMidiEvent.SaveToStream(s: TStream);
var b: integer;
    i: integer;
begin
  // Write Delta-Time
  for i:=0 to Time.Size-1 do
  begin
    s.Write(Time.Value[i], 1);
  end;

  // Write Event Number (Meta, Number, Channel)
  if Meta then
  begin
    b:=$FF; s.Write(b, 1);
    s.Write(Num, 1);
  end else begin
    b:=Num or Channel;
    s.Write(b, 1);
  end;
end;

{ TMidiEndOfTrackEvent }

constructor TMidiEndOfTrackEvent.Create;
begin
  inherited Create;
  Num:=$2F;
  Meta:=true;
end;

procedure TMidiEndOfTrackEvent.LoadFromStream(s: TStream);
var b: byte;
begin
  // Never used method (the track is never loaded, just created for saving)
  s.Read(b, 1);
end;

function TMidiEndOfTrackEvent.SaveSize: byte;
begin
  result:=time.Size+3;
end;

procedure TMidiEndOfTrackEvent.SaveToStream(s: TStream);
var b: byte;
begin
  inherited SaveToStream(s); // Time, Meta, Number, Channel
  b:=0; s.Write(b, 1); // Size (0)
end;

{ TMidiNoteONEvent }

constructor TMidiNoteONEvent.Create;
begin
  inherited Create;
  Num:=$90;
  Meta:=false;
end;

procedure TMidiNoteONEvent.LoadFromStream(s: TStream);
begin
  s.Read(Note, 1);
  s.Read(Velocity, 1);
end;

function TMidiNoteONEvent.SaveSize: byte;
begin
  result:=time.Size+3;
end;

procedure TMidiNoteONEvent.SaveToStream(s: TStream);
begin
  inherited SaveToStream(s); // Time, Meta, Number, Channel
  s.Write(Note, 1);
  s.Write(Velocity, 1);
end;

{ TMidiNoteOFFEvent }

constructor TMidiNoteOFFEvent.Create;
begin
  inherited Create;
  Num:=$80;
  Meta:=false;
end;

procedure TMidiNoteOFFEvent.LoadFromStream(s: TStream);
var b: byte;
begin
  s.Read(Note, 1);
  s.Read(b, 1); // Velocity isn't loaded
end;

function TMidiNoteOFFEvent.SaveSize: byte;
begin
  result:=time.Size+3;
end;

procedure TMidiNoteOFFEvent.SaveToStream(s: TStream);
var b: byte;
begin
  inherited SaveToStream(s); // Time, Meta, Number, Channel
  s.Write(Note, 1);
  b:=0; s.Write(b, 1); // Velocity
end;

{ TMidiChangePatchEvent }

constructor TMidiChangePatchEvent.Create;
begin
  inherited Create;
  Num:=$C0;
  Meta:=false;
end;

procedure TMidiChangePatchEvent.LoadFromStream(s: TStream);
begin
  s.Read(Patch, 1);
end;

function TMidiChangePatchEvent.SaveSize: byte;
begin
  result:=time.size+2;
end;

procedure TMidiChangePatchEvent.SaveToStream(s: TStream);
begin
  inherited SaveToStream(s); // Time, Meta, Number, Channel
  s.Write(Patch, 1);
end;

{ TMidi }

constructor TMidi.Create;
begin
  Tracks:=TList.Create;
end;

destructor TMidi.Destroy;
var i: integer;
begin
  for i:=0 to Tracks.Count-1 do TMidiTrack(Tracks[i]).Free;
  inherited Destroy;
end;

procedure TMidi.LoadFromFile(filename: string);
var s: TMemoryStream;
begin
  s:=TMemoryStream.Create;
  s.LoadFromFile(filename);
  LoadFromStream(s);
  s.Free;
end;

procedure TMidi.LoadFromStream(s: TStream);
var k: cardinal;
    w: word;
    error: boolean;
    trcnt: word;
    i: integer;
begin
  error:=false;
  s.Read(k, 4);
  if k<>$6468544D then error:=true; // MThd
  s.Read(k, 4);
  if k<>$06000000 then error:=true; // Size of header (always 6)

  if error then raise Exception.Create('Invalid MIDI file format.') else
  begin
    s.Read(w, 2); // Midi file format (2 bytes)
    w:=Invert2B(w);
    if w=0 then Format:=mfSingle else
    if w=2 then Format:=mfAsync else
      Format:=mfSync;
    s.Read(w, 2); // Tracks count
    trcnt:=Invert2B(w);
    s.Read(w, 2); // Number of delta-time ticks per quarter note
    TicksPerQNote:=Invert2B(w);
    // Tracks
    for i:=1 to trcnt do
    begin
      Tracks.Add(TMidiTrack.Create);
      TMidiTrack(Tracks[i-1]).LoadFromStream(s);
    end;
  end;
end;

procedure TMidi.SaveToFile(filename: string);
var s: TMemoryStream;
begin
  s:=TMemoryStream.Create;
  SaveToStream(s);
  s.SaveToFile(filename);
  s.Free;
end;

procedure TMidi.SaveToStream(s: TStream);
var //b: byte;
    w: word;
    i: integer;
begin
  s.Write('MThd', 4);
  w:=0; s.Write(w, 2); w:=Invert2B(6); s.Write(w, 2);
  if Format=mfSingle then w:=0 else
  if Format=mfAsync then w:=2 else
    w:=1;
  w:=Invert2B(w); s.Write(w, 2);
  w:=Invert2B(Tracks.Count); s.Write(w, 2);
  w:=Invert2B(TicksPerQNote); s.Write(w, 2);
  for i:=0 to Tracks.Count-1 do TMidiTrack(Tracks[i]).SaveToStream(s);
end;

{ TMidiOutput }

procedure TMidiOutput.ChangePatch(channel, patch: byte);
begin
  //$0000ppCc
  if channel>15 then raise Exception.Create('Invalid MIDI channel: '+inttostr(channel));
  ShortMsg($C0+(patch shl 8)+channel);
end;

procedure TMidiOutput.Close;
var res: integer;
begin
  if Opened then
  begin
    res:=midiOutClose(Handle);
    if res=MMSYSERR_NOERROR then
    begin
      FOpened:=false;
    end else begin
      raise Exception.Create('Couldn''t close MIDI output: error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
    end;
  end;
end;

constructor TMidiOutput.Create(ffDeviceID: integer; OpenNow: boolean);
begin
  FDeviceID:=ffDeviceID;
  FOpened:=false;
  if OpenNow then Open;
end;

destructor TMidiOutput.Destroy;
begin
  Close;
  inherited Destroy;
end;

procedure TMidiOutput.Drum(note, volume: byte);
begin
  NoteON(9, note, volume);
end;

function TMidiOutput.GetVolume: word;
var vol: cardinal;
    res: integer;
begin
  res:=midiOutGetVolume(Handle, @vol);
  if res=MMSYSERR_NOERROR then result:=trunc((LoWord(vol)+HiWord(vol))/2)
                          else raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

function TMidiOutput.GetVolumeLeft: word;
var vol: cardinal;
    res: integer;
begin
  res:=midiOutGetVolume(Handle, @vol);
  if res=MMSYSERR_NOERROR then result:=LoWord(vol)
                          else raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

function TMidiOutput.GetVolumeRight: word;
var vol: cardinal;
    res: integer;
begin
  res:=midiOutGetVolume(Handle, @vol);
  if res=MMSYSERR_NOERROR then result:=HiWord(vol)
                          else raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

procedure TMidiOutput.NoteOFF(channel, note: byte);
begin
  //$00vvnn8c
  if channel>15 then raise Exception.Create('Invalid MIDI channel: '+inttostr(channel));
  ShortMsg($80+(note shl 8)+channel);
end;

procedure TMidiOutput.NoteON(channel, note, volume: byte);
begin
  //$00vvnn9c
  if channel>15 then raise Exception.Create('Invalid MIDI channel: '+inttostr(channel));
  ShortMsg($90+(note shl 8)+(volume shl 16)+channel);
end;

procedure TMidiOutput.NotesOFF(channel: byte);
var i: integer;
begin
  for i:=0 to 127 do NoteOFF(channel, i);
end;

procedure TMidiOutput.Open;
var res: integer;
begin
  if Opened then Close;
  res:=midiOutOpen(@Handle, DeviceID, 0, 0, CALLBACK_NULL);
  if res=MMSYSERR_NOERROR then
  begin
    FOpened:=true;
  end else begin
    FOpened:=false;
    raise Exception.Create('Couldn''t open MIDI output: error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
  end;
end;

procedure TMidiOutput.SetDeviceID(const Value: integer);
begin
  Close;
  FDeviceID:=Value;
end;

procedure TMidiOutput.SetVolume(const Value: word);
var res: integer;
begin
  res:=midiOutSetVolume(Handle, Value+(Value shl 16));
  if res<>MMSYSERR_NOERROR then raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

procedure TMidiOutput.SetVolumeLeft(Value: word);
var res: integer;
begin
  res:=midiOutSetVolume(Handle, Value+(GetVolumeRight shl 16));
  if res<>MMSYSERR_NOERROR then raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

procedure TMidiOutput.SetVolumeRight(const Value: word);
var res: integer;
begin
  res:=midiOutSetVolume(Handle, GetVolumeLeft+(Value shl 16));
  if res<>MMSYSERR_NOERROR then raise Exception.Create('Midi error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
end;

procedure TMidiOutput.ShortMsg(msg: cardinal);
var res: integer;
begin
  if Opened then
  begin
    res:=midiOutShortMsg(Handle, msg);
    if res<>MMSYSERR_NOERROR then
    begin
      raise Exception.Create('Couldn''t send MIDI short message: error '+inttostr(res)+' ('+GetMidiErrorText(res)+')');
    end;
  end else begin
    raise Exception.Create('Couldn''t send MIDI short message: output is closed.');
  end;
end;

end.
